from .heart_disease_rf import get_heart_disease_rf
from .iris_lda import get_iris_test
from .ames_housing_gbm import get_ames_gbm_test
from .boston_gbm import get_boston_gbm_test
from .titanic_rf import get_titanic_rf
import data


from matplotlib import cm, colors, ticker
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np


class CiuObject:
    def __init__(self, ci, cu, c_mins, c_maxs, outval, intermediate_concepts, concept_names):
        self.ci = ci
        self.cu = cu
        self.c_mins = c_mins
        self.c_maxs = c_maxs
        self.outval = outval
        self.intermediate_concepts = intermediate_concepts
        self.concept_names = concept_names

    def _get_target_concept(self, target_concept, ind_inputs=None):
        # Initialising an ideal inputs copy including all the inputs of the concept
        ind_inputs_copy = []
        out_ci = self.ci.copy()
        for concept in self.intermediate_concepts:
            if target_concept in concept.keys():
                for feature_list in concept.values():
                    for feature in feature_list:
                        out_ci[feature] = (self.c_maxs[feature] - self.c_mins[feature]) / (
                                    self.c_maxs[target_concept] - self.c_mins[target_concept])
                        ind_inputs_copy.append(list(self.ci.keys()).index(feature))

        # Checking if there are user inputs, otherwise setting them to ideal copy
        if ind_inputs is None:
            ind_inputs = ind_inputs_copy

        # If there are inputs but outside of concept, toss a warning and set to ideal copy
        elif len(ind_inputs) >= 1:
            for i in ind_inputs:
                if i not in ind_inputs_copy:
                    print("WARNING: The indices selected must be a subset of the target concept.\n"
                          f"Index number {i} is outside of the concept scope.")
                    ind_inputs = ind_inputs_copy
                    break

        return out_ci, ind_inputs

    @staticmethod
    def _filter_feature_names(feature_names, concept_names, include_intermediate_concepts, ind_inputs=None):
        feature_names_final = []
        original = [x for x in feature_names if x not in concept_names]
        exclude_intermediate_concepts = include_intermediate_concepts == 'no'
        exclude_ordinary = include_intermediate_concepts == 'only'
        for feature_name in feature_names:
            exclude_feature = \
                (exclude_intermediate_concepts and feature_name in concept_names) or \
                (exclude_ordinary and feature_name not in concept_names)
            if not exclude_feature:
                if exclude_ordinary:
                    feature_names_final.append(feature_name)
                if not exclude_ordinary and ind_inputs is None:
                    feature_names_final.append(feature_name)
            if ind_inputs:
                feature_names_final.extend(
                    list(original)[index] for index in ind_inputs if list(original)[index] not in feature_names_final)
        return feature_names_final

    def explain_tabular(self, include_intermediate_concepts=None, ind_inputs=None, target_concept=None):
        """
        :param str target_concept: defines which intermediate concept to explain;
        :param list ind_inputs: list of feature indexes to produce a tabular explanation for, it will include all of them by default;
                                NOTE: this can add extra indexes to explain even if the include_intermediate_concepts param is set to 'only'
        :param str include_intermediate_concepts: define whether to include 'no' intermediate concepts or 'only' intermediate concepts;
                                                    it will include all intermediate concepts and all independent features by default
        :return: dataframe output_df: Pandas dataframe containing the output data
        """
        if target_concept:
            # Removing the inclusion of the other concepts automatically
            include_intermediate_concepts = None
            out_ci, ind_inputs = self._get_target_concept(
                target_concept, ind_inputs
            )

        if target_concept is None:
            out_ci = self.ci

        out_df = pd.DataFrame.from_dict([out_ci,
                                         self.cu,
                                         self.c_mins,
                                         self.c_maxs])

        out_df.index = pd.Index(['CI', 'CU', 'cmin', 'cmax'], name='Features')

        feature_names_out = self._filter_feature_names(
            self.ci.keys(),
            self.concept_names,
            include_intermediate_concepts,
            ind_inputs
        )

        out_df = out_df[feature_names_out]

        output_df = out_df.T
        output_df['outval'] = list(self.outval.values())[0]

        return output_df

    def explain_text(self, include_intermediate_concepts=None, ind_inputs=None, thresholds_ci=None, thresholds_cu=None, target_concept=None):
        """
        :param str target_concept: defines which intermediate concept to explain;
        :param list ind_inputs: list of feature indexes to produce a textual explanation for, it will include all of them by default;
                                NOTE: this can add extra indexes to explain even if the include_intermediate_concepts param is set to 'only'
        :param str include_intermediate_concepts: define whether to include 'no' intermediate concepts or 'only' intermediate concepts;
                                                    it will include all intermediate concepts and all independent features by default
        :param dict thresholds_ci: dictionary containing the label and ceiling value for the CI thresholds
        :param dict thresholds_cu: dictionary containing the label and ceiling value for the CU thresholds
        :return: list explanation_texts: list containing explanation strings
        """

        if thresholds_ci is None:
            thresholds_ci = {
                'very low importance': 0.20,
                'low importance': 0.40,
                'normal importance': 0.60,
                'high importance': 0.80,
                'very high importance': 1
            }

        if thresholds_cu is None:
            thresholds_cu = {
                'not typical': 0.25,
                'somewhat typical': 0.5,
                'typical': 0.75,
                'very typical': 1
            }

        if len(thresholds_cu) < 2 or len(thresholds_ci) < 2:
            raise ValueError(f"The dictionaries containing the CI/CU thresholds must have at least 2 elements. \
                             \nCI dict: {thresholds_ci} \nCU dict: {thresholds_cu}")

        if target_concept:
            #Removing the inclusion of the other concepts automatically
            include_intermediate_concepts = None
            out_ci, ind_inputs = self._get_target_concept(
                target_concept, ind_inputs
            )

        if target_concept is None:
            out_ci = self.ci

        feature_names = self._filter_feature_names(
            self.ci.keys(),
            self.concept_names,
            include_intermediate_concepts,
            ind_inputs
        )

        explanation_texts = []

        if target_concept:
            for k, v in thresholds_cu.items():
                if self.cu[target_concept] <= v:
                    cu_concept_text = k
                    break

            cu_concept = round(self.cu[target_concept] * 100, 2)

            concept_text = f'The intermediate concept "{target_concept}", is {cu_concept_text} ' \
                                       f'for its prediction (CU={cu_concept}%).'
            explanation_texts.append(concept_text)

        for feature in list(feature_names):
            try:
                for k, v in thresholds_ci.items():
                    if out_ci[feature] <= v:
                        ci_text = k
                        break

                for k, v in thresholds_cu.items():
                    if self.cu[feature] <= v:
                        cu_text = k
                        break
            except TypeError as e:
                raise TypeError(f"The dictionaries containing the CI/CU thresholds cannot have \x1B[3mNone\x1B[0m values. \
                                    \nCI dict: {thresholds_ci} \nCU dict: {thresholds_cu}") from e


            ci = round(out_ci[feature] * 100, 2)
            cu = round(self.cu[feature] * 100, 2)

            explanation_text = f'The feature "{feature}", which is of ' \
                                       f'{ci_text} (CI={ci}%), is {cu_text} ' \
                                       f'for its prediction (CU={cu}%).'

            explanation_texts.append(explanation_text)

        return explanation_texts

    def plot_ciu(self, plot_mode='default', include_intermediate_concepts=None,
                 ind_inputs=None, target_concept=None, sort='ci', color_blind=None,
                 color_fill_ci='#7fffd44d', color_edge_ci='#66CDAA',
                 color_fill_cu="#006400cc", color_edge_cu="#006400"):

        """
        :param str target_concept: defines which intermediate concept to explain;
        :param list ind_inputs: list of feature indexes to produce a plot explanation for, it will include all of them by default;
                                NOTE: this can add extra indexes to explain even if the include_intermediate_concepts param is set to 'only'
        :param str include_intermediate_concepts: define whether to include 'no' intermediate concepts or 'only' intermediate concepts;
                                                    otherwise it will include all intermediate concepts and all independent features by default
        :param str plot_mode: defines the type plot to use between 'default', 'overlap' and 'combined'.
        :param str sort: defines the order of the plot bars by the 'ci' (default), 'cu' values or unsorted if None;
        :param str color_blind: defines accessible color maps to use for the plots, such as 'protanopia',
                                                        'deuteranopia' and 'tritanopia'.
        :param str color_edge_cu: defines the hex or named color for the CU edge in the overlap plot mode.
        :param str color_fill_cu: defines the hex or named color for the CU fill in the overlap plot mode.
        :param str color_edge_ci: defines the hex or named color for the CI edge in the overlap plot mode.
        :param str color_fill_ci: defines the hex or named color for the CI fill in the overlap plot mode.
        """

        if target_concept:
            # Removing the inclusion of the other concepts automatically
            include_intermediate_concepts = None
            out_ci, ind_inputs = self._get_target_concept(
                target_concept, ind_inputs
            )

        if target_concept is None:
            out_ci = self.ci

        data = np.fromiter(out_ci.values(), dtype=float)
        cu = np.fromiter(self.cu.values(), dtype=float)

        fig, ax = plt.subplots(figsize=(6, 6))

        if target_concept:
            fig.suptitle(f"The target concept is {target_concept}")

        feature_names_prelim = self.ci.keys()
        feature_names = self._filter_feature_names(
            feature_names_prelim,
            self.concept_names,
            include_intermediate_concepts,
            ind_inputs
        )

        indices_deleted = 0
        for index, feature_name in enumerate(feature_names_prelim):
            if feature_name not in feature_names:
                data = np.delete(data, index - indices_deleted)
                cu = np.delete(cu, index - indices_deleted)
                indices_deleted += 1

        y_pos = np.arange(len(feature_names))

        if sort == 'ci':
            data, cu, feature_names = (list(t) for t in zip(*sorted(zip(data, cu, feature_names))))
        elif sort == 'cu':
            cu, data, feature_names = (list(t) for t in zip(*sorted(zip(cu, data, feature_names))))

        my_norm = colors.Normalize(vmin=0, vmax=1)
        nodes = [0.0, 0.5, 1.0]

        if color_blind is None:
            colours = ["red", "yellow", "green"]
        elif color_blind == 'protanopia':
            colours = ["gray", "yellow", "blue"]
        elif color_blind == 'deuteranopia':
            colours = ["slategray", "orange", "dodgerblue"]
        elif color_blind == 'tritanopia':
            colours = ["#ff0066", "#ffe6f2", "#00e6e6"]

        cmap1 = colors.LinearSegmentedColormap.from_list("mycmap", list(zip(nodes, colours)))

        sm = cm.ScalarMappable(cmap=cmap1, norm=my_norm)
        sm.set_array([])

        if plot_mode == "default":
            cbar = plt.colorbar(sm, ax=plt.gca())
            cbar.set_label('CU', rotation=0, labelpad=25)

            plt.xlabel("CI")

            for m in range(len(data)):
                ax.barh(y_pos[m], data[m], color=cmap1(my_norm(cu[m])),
                        edgecolor="#808080", zorder=2)

        if plot_mode == "overlap":
            plt.xlabel("CI and relative CU")

            for m in range(len(data)):
                ax.barh(y_pos[m], data[m], color=color_fill_ci,
                        edgecolor=color_edge_ci, linewidth=1.5, zorder=2)
                ax.barh(y_pos[m], cu[m] * data[m], color=color_fill_cu,
                        edgecolor=color_edge_cu, linewidth=1.5, zorder=2)

        if plot_mode == "combined":
            plt.xlabel("CI and relative CU")

            cbar = plt.colorbar(sm, ax=plt.gca())
            cbar.set_label('CU', rotation=0, labelpad=25)

            for m in range(len(data)):
                ax.barh(y_pos[m], data[m], color="#ffffff66", edgecolor="#808080", zorder=2)
                ax.barh(y_pos[m], cu[m] * data[m], color=cmap1(my_norm(cu[m])), zorder=2)

        ax.set_facecolor(color="#D9D9D9")
        ax.set_xlim(0, 1)

        ax.set_yticks(y_pos)
        ax.set_yticklabels(feature_names)
        ax.set_xticks([0.00, 0.25, 0.50, 0.75, 1.00])
        ax.tick_params(axis='x', labelsize=12)
        ax.tick_params(axis='y', labelsize=12)

        ax.xaxis.set_minor_locator(ticker.MultipleLocator(0.25 / 2))
        ax.yaxis.set_minor_locator(ticker.MultipleLocator(1))
        ax.grid(which='minor')
        ax.grid(which='minor', color='white')
        ax.grid(which='major', color='white')

        plt.ylabel("Features")

        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['bottom'].set_visible(False)
        ax.spines['left'].set_visible(True)



import os


def get_titanic_rf():
    """
    Random forest model using the intermediate concepts on the Titanic dataset
    :return: Titanic CIU object with intermediate concepts
    """

    import pandas as pd
    from sklearn.ensemble import RandomForestClassifier
    from ciu.ciu_core import determine_ciu

    data = pd.read_csv(os.path.join(os.path.dirname(__file__), "titanic.csv"))

    data = data.drop(data.columns[0], axis=1)
    unused = ['PassengerId','Cabin','Name','Ticket']

    for col in unused:
        data = data.drop(col, axis=1)

    from sklearn.preprocessing import LabelEncoder
    data = data.dropna().apply(LabelEncoder().fit_transform)
    train = data.drop('Survived', axis=1)

    new_passenger = pd.DataFrame.from_dict({"Pclass" : [1], "Sex": [1], "Age": [8], "SibSp": [0], "Parch": [0], "Fare": [72], "Embarked": [2]})

    model = RandomForestClassifier(n_estimators=100)
    model.fit(train, data.Survived)

    intermediate_tit = [
            {"Wealth":['Pclass', 'Fare']},
            {"Family":['SibSp', 'Parch']},
            {"Gender":['Sex']},
            {"Age_years":['Age']},
            {"Embarked_Place":['Embarked']}
        ]

    ciu_tit = determine_ciu(
        new_passenger,
        model.predict_proba,
        train.to_dict('list'),
        samples = 1000,
        prediction_index = 1,
        intermediate_concepts=intermediate_tit
    )

    return ciu_tit



def get_ames_gbm_test():
    """
    :return: A CIU object and the list of intermediate concepts used in the example.
    """
    from ciu.ciu_core import determine_ciu
    import pandas as pd
    import xgboost as xgb
    from sklearn.model_selection import train_test_split
    import os

    df = pd.read_csv(os.path.join(os.path.dirname(__file__), "AmesHousing.csv"))

    #Checking for missing data
    missing_data_count = df.isnull().sum()
    missing_data_percent = df.isnull().sum() / len(df) * 100

    missing_data = pd.DataFrame({
        'Count': missing_data_count,
        'Percent': missing_data_percent
    })

    missing_data = missing_data[missing_data.Count > 0]
    missing_data.sort_values(by='Count', ascending=False, inplace=True)

    #This one has spaces for some reason
    df.columns = df.columns.str.replace(' ', '')


    #Taking care of missing values
    from sklearn.impute import SimpleImputer
    # Group 1:
    group_1 = [
        'PoolQC', 'MiscFeature', 'Alley', 'Fence', 'FireplaceQu', 'GarageType',
        'GarageFinish', 'GarageQual', 'GarageCond', 'BsmtQual', 'BsmtCond',
        'BsmtExposure', 'BsmtFinType1', 'BsmtFinType2', 'MasVnrType'
    ]
    df[group_1] = df[group_1].fillna("None")

    # Group 2:
    group_2 = [
        'GarageArea', 'GarageCars', 'BsmtFinSF1', 'BsmtFinSF2', 'BsmtUnfSF',
        'TotalBsmtSF', 'BsmtFullBath', 'BsmtHalfBath', 'MasVnrArea'
    ]

    df[group_2] = df[group_2].fillna(0)

    # Group 3:
    group_3a = [
        'Functional', 'MSZoning', 'Electrical', 'KitchenQual', 'Exterior1st',
        'Exterior2nd', 'SaleType', 'Utilities'
    ]

    imputer = SimpleImputer(strategy='most_frequent')
    df[group_3a] = pd.DataFrame(imputer.fit_transform(df[group_3a]), index=df.index)

    df.LotFrontage = df.LotFrontage.fillna(df.LotFrontage.mean())
    df.GarageYrBlt = df.GarageYrBlt.fillna(df.YearBuilt)

    #Label encoding
    from sklearn.preprocessing import LabelEncoder
    df = df.apply(LabelEncoder().fit_transform)

    data = df.drop(columns=['SalePrice'])
    target = df.SalePrice

    #Splitting and training
    X_train, X_test, y_train, y_test = train_test_split(data, target, test_size=0.3, random_state=123)
    xg_reg = xgb.XGBRegressor(objective ='reg:squarederror', colsample_bytree = 0.5, learning_rate = 0.1, max_depth = 15, alpha = 10)

    xg_reg.fit(X_train,y_train)


    intermediate = [
        {"Garage":[c for c in df.columns if 'Garage' in c]},
        {"Basement":[c for c in df.columns if 'Bsmt' in c]},
        {"Lot":list(df.columns[[3,4,7,8,9,10,11]])},
        {"Access":list(df.columns[[13,14]])},
        {"House_type":list(df.columns[[1,15,16,21]])},
        {"House_aesthetics":list(df.columns[[22,23,24,25,26]])},
        {"House_condition":list(df.columns[[20,18,21,28,19,29]])},
        {"First_floor_surface":list(df.columns[[43]])},
        {"Above_ground_living area":[c for c in df.columns if 'GrLivArea' in c]}
    ]


    test_data_ames = X_test.iloc[[345]]

    ciu = determine_ciu(
        test_data_ames,
        xg_reg.predict,
        df.to_dict('list'),
        samples = 1000,
        prediction_index = None,
        intermediate_concepts = intermediate
    )

    return ciu, intermediate



def get_iris_test():
    """
    :return: iris LDA CIU Object
    """

    from sklearn.model_selection import train_test_split
    from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
    from sklearn import datasets
    import pandas as pd
    import numpy as np
    from ciu.ciu_core import determine_ciu

    iris=datasets.load_iris()

    df = pd.DataFrame(data = np.c_[iris['data'], iris['target']],
                  columns = iris['feature_names'] + ['target'])
    df['species'] = pd.Categorical.from_codes(iris.target, iris.target_names)
    df.columns = ['s_length', 's_width', 'p_length', 'p_width', 'target', 'species']

    X = df[['s_length', 's_width', 'p_length', 'p_width']]
    y = df['species']
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.3, random_state=123)


    model = LinearDiscriminantAnalysis()
    model.fit(X_train, y_train)

    iris_df = df.apply(pd.to_numeric, errors='ignore')

    #Can also be written manually:
    #   test_iris = pd.DataFrame.from_dict({'s_length' : [2], 's_width' : [3.2], 'p_length': [1.8], 'p_width' : [2.4]})

    ciu = determine_ciu(
        X_test.iloc[[42]],
        model.predict_proba,
        iris_df.to_dict('list'),
        samples = 1000,
        prediction_index = 2
    )

    return ciu



def get_heart_disease_rf():
    """
    :return: heart disease CIU Object with a Random Forest Classifier
    """
    from sklearn.ensemble import RandomForestClassifier
    from sklearn.model_selection import train_test_split
    import pandas as pd
    import numpy as np
    from ciu.ciu_core import determine_ciu

    model = RandomForestClassifier(n_estimators=100)

    df = pd.read_csv("https://archive.ics.uci.edu/ml/machine-learning-databases/heart-disease/processed.cleveland.data")
    df.columns = ["age", "sex", "cp", "trestbps", "chol","fbs", "restecg",
                          "thalach","exang", "oldpeak","slope", "ca", "thal", "num"]

    df = df.replace({'?':np.nan}).dropna()

    df.loc[df["num"] > 0, "num"] = 1

    X = df.drop('num',axis=1)
    y = df['num']

    for i in df.columns:
        if 'object' in str(df[i].dtypes):
            df[i] = df[i].astype(float)

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.3, random_state=123)

    model.fit(X_train, y_train)

    ciu = determine_ciu(
        X_test.iloc[[23]],
        model.predict_proba,
        df.to_dict('list'),
        samples = 1000,
        prediction_index = 0
    )

    return ciu


import pandas as pd
pd.options.mode.chained_assignment = None
import random
from ciu.ciu_object import CiuObject
import numpy as np


def _generate_samples(case, feature_names, min_maxs, samples, indices,
                      category_mapping):
    rows = [case]
    for _ in range(samples):
        sample_entry = {}
        for index_j, feature_j in enumerate(feature_names):
            if index_j not in indices:
                # if not (index_j in indices):
                sample_entry[feature_j] = case[feature_j]
            else:
                min_val = min_maxs[feature_j][0]
                max_val = min_maxs[feature_j][1]
                sample_entry[feature_j] = random.uniform(min_val, max_val)
                # check if feature_j, feature_k in same category;
                # if so, set feature_k to 0 if feature_j is 1
                for index_k, feature_k in enumerate(feature_names):
                    if index_j != index_k:
                        for categories in category_mapping.values():
                            same_category = feature_j in categories \
                                            and feature_k in categories
                            if same_category and sample_entry[feature_j] == 1:
                                sample_entry[feature_k] = 0

        # set categorical values that would otherwise not have a category
        # assigned
        for categories in category_mapping.values():
            is_activated = False
            for category in categories:
                try:
                    if (sample_entry[category] == 1).any(): is_activated = True
                except AttributeError:
                    if sample_entry[category] == 1: is_activated = True
            if not is_activated:
                category = categories[random.randint(0, len(categories) - 1)]
                sample_entry[category] = 1
        rows.append(sample_entry)

    return rows


def determine_ciu(
        case, predictor, dataset=None, min_maxs=None, samples=1000,
        prediction_index=None, category_mapping=None, intermediate_concepts=None):
    """
    Determines contextual importance and utility for a given case.

    :param case: Case data (Pandas DataFrame)
    :param predictor: The prediction function of the black-box model Py-CIU should call
    :param dict dataset: Dataset to deduct min_maxs from (dictionary).
                            Defaults to ``None``.
    :param min_maxs: dictionary (``'feature_name': [min, max, is_int]`` for each feature),
                    or infered from dataset. Defaults to ``None``
    :param samples: number of samples to be generated. Defaults to 1000.
    :param prediction_index: If the model returns several predictions, provide
                             the index of the relevant prediction. Defaults to
                             ``None``
    :param category_mapping: Mapping of one-hot encoded categorical variables to
                             list of categories and category name. Defaults to
                             ``None``.
    :param intermediate_concepts: List of {key: list} tuples of features whose
                                 interactions should be evaluated. Defaults to
                                 ``[]``.

    :return: dictionary: for each feature: list with
             contextual importance value, contextual utility value
    """

    if intermediate_concepts is None:
        intermediate_concepts = []

    if category_mapping is None:
        category_mapping = {}

    for i in case.columns:
        case[i] = case[i].astype(float)

    if min_maxs is None:
        try:
            min_maxs = {i: [min(list(map(float, dataset[i]))), max(list(map(float, dataset[i]))),
                            isinstance(min(dataset[i]), float)] for i in case.columns}
        except:
            print("Logic Error: You must provide either min_max values or a dataset to infer them from")
            raise

    category_names = list(category_mapping.keys())
    feature_names_decoded = []
    categories_encoded = []
    for feature in min_maxs.keys():
        is_category = False
        for index, categories in enumerate(category_mapping.values()):
            if feature in categories:
                categories_encoded.append(feature)
                is_category = True
                if category_names[index] not in feature_names_decoded:
                    feature_names_decoded.append(category_names[index])
        if not is_category:
            feature_names_decoded.append(feature)

    cis = {}
    cus = {}
    c_mins = {}
    c_maxs = {}
    outval = {'outval': predictor(case)[0][prediction_index] if prediction_index is not None else predictor(case)[0]}

    case_prediction = \
        predictor(case)[0] if prediction_index is None \
            else predictor(case)[0][prediction_index]

    all_samples = []

    for index_i, feature_i in enumerate(min_maxs.keys()):
        feature_samples = _generate_samples(
            case.iloc[0, :].to_dict(), min_maxs.keys(), min_maxs, samples, [index_i],
            category_mapping
        )
        all_samples.append(feature_samples)

    joined_samples = [j for i in all_samples for j in i]

    try:
        all_samples_flat = pd.DataFrame(joined_samples)
        all_preds = predictor(all_samples_flat)
    except ValueError:
        all_samples_flat = pd.DataFrame(joined_samples, dtype=float)
        all_preds = predictor(all_samples_flat)

    predictions = {feat: all_preds[ind * samples + ind: (ind + 1) * samples + ind + 1] if prediction_index is None else \
        [prob[prediction_index] for prob in all_preds[ind * samples + ind: (ind + 1) * samples + ind + 1]] \
                   for ind, feat in enumerate(min_maxs.keys())}

    for intermediate_concept in intermediate_concepts:
        interaction_name = list(intermediate_concept.keys())[0]
        features = list(intermediate_concept.values())[0]
        indices = [list(min_maxs.keys()).index(feature) for feature in features]

        feature_samples = _generate_samples(
            case.iloc[0, :].to_dict(), min_maxs.keys(), min_maxs, samples, indices, category_mapping
        )

        try:
            feature_samples = pd.DataFrame(feature_samples)
            predictions[interaction_name] = \
                predictor(feature_samples) if prediction_index is None \
                    else [prob[prediction_index] for \
                          prob in predictor(feature_samples)]
        except ValueError:
            feature_samples = pd.DataFrame(feature_samples, dtype=float)
            predictions[interaction_name] = \
                predictor(feature_samples) if prediction_index is None \
                    else [prob[prediction_index] for \
                          prob in predictor(feature_samples)]

    abs_max = None
    abs_min = None

    # determine absolute min/max, only considering single features
    for feature in feature_names_decoded:
        # Get right predictions for decoded category
        if feature in category_mapping.keys():
            encoded_feature = None
            for encoded_feature_j in min_maxs.keys():
                feature_max = max(predictions[encoded_feature_j])
                if abs_max is None or abs_max < feature_max:
                    abs_max = feature_max
                feature_min = min(predictions[encoded_feature_j])
                if abs_min is None or abs_min > feature_min:
                    abs_min = feature_min
        else:
            feature_max = max(predictions[feature])
            if abs_max is None or abs_max < feature_max:
                abs_max = feature_max

            feature_min = min(predictions[feature])
            if abs_min is None or abs_min > feature_min:
                abs_min = feature_min

    # determine absolute min/max, also considering feature interactions
    for intermediate_concept in intermediate_concepts:
        intermediate_concept_name = list(intermediate_concept.keys())[0]
        interaction_max = max(predictions[intermediate_concept_name])
        if abs_max is None or abs_max < interaction_max:
            abs_max = interaction_max

        interaction_min = min(predictions[intermediate_concept_name])
        if abs_min is None or abs_min > interaction_min:
            abs_min = interaction_min

    # compute CI/CU for single features
    for index, feature in enumerate(feature_names_decoded):
        # Get right predictions for decoded category
        if feature in category_mapping.keys():
            encoded_feature = None
            for encoded_feature_j in min_maxs.keys():
                if encoded_feature_j in category_mapping[feature] and list(case[encoded_feature_j].to_dict().values())[0] == 1:
                    encoded_feature = encoded_feature_j
            c_min = min(predictions[encoded_feature])
            c_max = max(predictions[encoded_feature])
        else:
            c_min = min(predictions[feature])
            c_max = max(predictions[feature])

        n = case_prediction
        ci = (c_max - c_min) / (abs_max - abs_min)
        if (c_max - c_min) == 0:
            cu = (n - c_min) / 0.01
        else:
            cu = (n - c_min) / (c_max - c_min)
        if cu == 0: cu = 0.001
        cis[feature] = ci
        cus[feature] = cu
        c_mins[feature] = c_min
        c_maxs[feature] = c_max

    # compute CI/CU for feature interactions
    intermediate_concept_names = [
        "_".join(features) for features in intermediate_concepts
    ]

    for intermediate_concept_name in intermediate_concept_names:
        c_min = min(predictions[intermediate_concept_name])
        c_max = max(predictions[intermediate_concept_name])
        n = case_prediction
        ci = (c_max - c_min) / (abs_max - abs_min)
        if (c_max - c_min) == 0:
            cu = (n - c_min) / 0.01
        else:
            cu = (n - c_min) / (c_max - c_min)
        if cu == 0: cu = 0.001
        cis[intermediate_concept_name] = ci
        cus[intermediate_concept_name] = cu
        c_mins[intermediate_concept_name] = c_min
        c_maxs[intermediate_concept_name] = c_max

    return CiuObject(cis, cus, c_mins, c_maxs, outval, intermediate_concepts, intermediate_concept_names)



import pathlib
import sys

from setuptools import setup, find_packages

long_desc = pathlib.Path('README.md').read_text()

if sys.version_info < (3, 6):
    print('CIU requires at least Python 3.6 due to the use of f-strings.')

setup(
    name='py-ciu',
    version='0.1.0.1',
    install_requires=[
        "matplotlib",
        "numpy",
        "pandas",
        "sklearn",
        "xgboost",
        "scikit_learn"
    ],
    url='https://github.com/KaryFramling/py-ciu',
    license='MIT',
    author='Vlad Apopei & Kary Främling',
    author_email='ioan-vlad.apopei@aalto.fi, kary.framling@umu.se',
    description='Python documentation generator',
    long_description=long_desc,
    long_description_content_type='text/markdown',
    project_urls={
        "Code": "https://github.com/KaryFramling/py-ciu",
    },
    platforms='any',
    packages=find_packages(include=['ciu', 'ciu_tests', 'ciu_tests', 'ciu_tests.data']),
    package_data={
        '': ['*.csv'],
    },
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Topic :: Documentation",
    ],
)




def get_boston_gbm_test():
    """
    :return: boston gbm CIU Object
    """

    import numpy as np
    import pandas as pd
    import xgboost as xgb
    from ciu.ciu_core import determine_ciu
    from sklearn.model_selection import train_test_split

    data_url = "http://lib.stat.cmu.edu/datasets/boston"
    raw_df = pd.read_csv(data_url, sep="\s+", skiprows=22, header=None)
    data = np.hstack([raw_df.values[::2, :], raw_df.values[1::2, :2]])
    target = raw_df.values[1::2, 2]

    data = pd.DataFrame(data)
    data.columns = ['CRIM', 'ZN', 'INDUS', 'CHAS', 'NOX', 'RM', 'AGE', 'DIS', 'RAD', 'TAX', 'PTRATIO', 'B', 'LSTAT']

    xgb.DMatrix(data=data,label=target)

    X_train, X_test, y_train, y_test = train_test_split(data, target, test_size=0.3, random_state=123)
    xg_reg = xgb.XGBRegressor(objective ='reg:squarederror', colsample_bytree = 0.3, learning_rate = 0.1,max_depth = 5, alpha = 10, n_estimators = 10)

    xg_reg.fit(X_train,y_train)

    ciu = determine_ciu(
        X_test.iloc[[124]],
        xg_reg.predict,
        data.to_dict('list'),
        samples = 1000,
        prediction_index = None
    )

    return ciu
